# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-16  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import six
import logging
import os
from six.moves.urllib.request import urlopen
import xml.etree.ElementTree as ET

import keyring
import requests
from requests_oauthlib import OAuth2Session

from .configstore import config_store, key_store
from .descriptive import MultiLineEdit
from .pyqt import Busy, Qt, QtCore, QtGui, QtWidgets
from .uploader import PhotiniUploader

logger = logging.getLogger(__name__)

nsmap = {
    'atom'   : 'http://www.w3.org/2005/Atom',
    'gd'     : 'http://schemas.google.com/g/2005',
    'georss' : 'http://www.georss.org/georss',
    'gml'    : 'http://www.opengis.net/gml',
    'gphoto' : 'http://schemas.google.com/photos/2007',
    'media'  : 'http://search.yahoo.com/mrss/',
    }

for prefix, uri in nsmap.items():
    if prefix == 'atom':
        prefix = ''
    ET.register_namespace(prefix, uri)


# general node, can be any kind of element
class PicasaNode(object):
    _elements = {
        # name         namespace repeat node
        'access'    : ('gphoto', False, False),
        'category'  : ('atom',   False, False),
        'group'     : ('media',  False, True),
        'id'        : ('gphoto', False, False),
        'keywords'  : ('media',  False, False),
        'location'  : ('gphoto', False, False),
        'numphotos' : ('gphoto', False, False),
        'Point'     : ('gml',    False, True),
        'pos'       : ('gml',    False, False),
        'summary'   : ('atom',   False, False),
        'thumbnail' : ('media',  True,  False),
        'timestamp' : ('gphoto', False, False),
        'title'     : ('atom',   False, False),
        'where'     : ('georss', False, True),
        }

    def __init__(self, dom=None, text=None):
        if text is not None:
            self._dom = ET.fromstring(text.encode('utf-8'))
        elif dom is not None:
            self._dom = dom
        else:
            self._dom = ET.Element('{' + nsmap['atom'] + '}entry')
        self.etag = self._dom.get('{' + nsmap['gd'] + '}etag')

    def __getattr__(self, name):
        if name not in self._elements:
            raise AttributeError(name)
        namespace, repeat, node = self._elements[name]
        tag = '{' + nsmap[namespace] + '}' + name
        if repeat:
            result = self._dom.findall(tag)
            if node:
                result = list(map(PicasaNode, result))
            return result
        result = self._dom.find(tag)
        if result is None:
            logger.debug('new %s', name)
            result = ET.SubElement(self._dom, tag)
        if node:
            result = PicasaNode(result)
        return result

    def to_string(self):
        return ET.tostring(self._dom, encoding='utf-8')

    def get_link(self, link_type):
        for child in self._dom.findall('atom:link', nsmap):
            if child.get('rel').endswith(link_type):
                return child.get('href')
        return None


# special node for the album feed, to disambiguate 'thumbnail'
class FeedNode(PicasaNode):
    _elements = {
        # name         namespace repeat node
        'entry'     : ('atom',   True,  True),
        'thumbnail' : ('gphoto', False, False),
        'nickname'  : ('gphoto', False, False),
        }


class PicasaSession(object):
    token_url     = 'https://www.googleapis.com/oauth2/v4/token'
    album_feed    = 'https://picasaweb.google.com/data/feed/api/user/default'
    scope = {
        'read' : 'https://picasaweb.google.com/data/',
        'write': 'https://picasaweb.google.com/data/',
        }

    def __init__(self, auto_refresh=True):
        self.auto_refresh = auto_refresh
        self.session = None
        self.token = None

    def log_out(self):
        keyring.delete_password('photini', 'picasa')
        self.session = None

    def permitted(self, level):
        refresh_token = keyring.get_password('photini', 'picasa')
        if not refresh_token:
            self.session = None
            self.token = None
            return False
        if not self.token:
            # create expired token
            self.token = {
                'access_token'  : 'xxx',
                'refresh_token' : refresh_token,
                'expires_in'    : -30,
                }
            self.session = None
        if not self.session:
            # create new session
            client_id     = key_store.get('picasa', 'client_id')
            client_secret = key_store.get('picasa', 'client_secret')
            auto_refresh_kwargs = {
                'client_id'    : client_id,
                'client_secret': client_secret,
                }
            if self.auto_refresh:
                self.session = OAuth2Session(
                    client_id, token=self.token, token_updater=self._save_token,
                    auto_refresh_kwargs=auto_refresh_kwargs,
                    auto_refresh_url=self.token_url,
                    )
            else:
                self.session = OAuth2Session(client_id, token=self.token)
            # refresh manually to get a valid token now
            self.token = self.session.refresh_token(
                self.token_url, **auto_refresh_kwargs)
            self.session.headers.update({'GData-Version': 2})
            # verify the token
            resp = self._check_response(self.session.get(
                'https://www.googleapis.com/oauth2/v3/tokeninfo',
                params={'access_token': self.token['access_token']})).json()
            if resp['scope'] != self.scope[level] or resp['aud'] != client_id:
                return False
        return True

    def get_auth_url(self, level):
        logger.info('using %s', keyring.get_keyring().__module__)
        client_id = key_store.get('picasa', 'client_id')
        self.session = OAuth2Session(
            client_id, scope=self.scope[level],
            redirect_uri='urn:ietf:wg:oauth:2.0:oob')
        return self.session.authorization_url(
            'https://accounts.google.com/o/oauth2/v2/auth')[0]

    def get_access_token(self, auth_code):
        # Fix for requests-oauthlib bug #157
        # https://github.com/requests/requests-oauthlib/issues/157
        os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = 'True'
        client_id = key_store.get('picasa', 'client_id')
        client_secret = key_store.get('picasa', 'client_secret')
        self.token = self.session.fetch_token(
            self.token_url, code=auth_code,
            auth=requests.auth.HTTPBasicAuth(client_id, client_secret))
        self._save_token(self.token)
        self.session = None

    def _save_token(self, token):
        if self.auto_refresh:
            keyring.set_password('photini', 'picasa', token['refresh_token'])

    def edit_node(self, node):
        resp = self._check_response(self.session.put(
            node.get_link('edit'), node.to_string(),
            headers={'If-Match' : node.etag,
                     'Content-Type' : 'application/atom+xml'}))
        return PicasaNode(text=resp.text)

    def get_feed(self):
        resp = self._check_response(self.session.get(
            self.album_feed, params={'kind': 'album'}))
        return FeedNode(text=resp.text)

    def new_album(self, album):
        resp = self._check_response(self.session.post(
            self.album_feed, album.to_string(),
            headers={'Content-Type' : 'application/atom+xml'}))
        return PicasaNode(text=resp.text)

    def delete_album(self, album):
        self._check_response(self.session.delete(
            album.get_link('edit'), headers={'If-Match' : album.etag}))

    def new_photo(self, title, album, data, image_type):
        resp = self._check_response(self.session.post(
            album.get_link('feed'), data=data,
            headers={'Content-Type' : 'image/' + image_type, 'Slug' : title}))
        return PicasaNode(text=resp.text)

    def do_upload(self, fileobj, image_type, image, params):
        # upload photo
        title = os.path.basename(image.path)
        try:
            photo = self.new_photo(title, params, fileobj, image_type)
        except Exception as ex:
            return str(ex)
        # set metadata
        photo.title.text = title
        title = image.metadata.title
        description = image.metadata.description
        if title and description:
            photo.summary.text = title.value + '\n\n' + description.value
        elif title:
            photo.summary.text = title.value
        elif description:
            photo.summary.text = description.value
        keywords = image.metadata.keywords
        if keywords:
            photo.group.keywords.text = ','.join(keywords.value)
        latlong = image.metadata.latlong
        if latlong:
            photo.where.Point.pos.text = '{:.6f} {:.6f}'.format(
                latlong.lat, latlong.lon)
        try:
            self.edit_node(photo)
        except Exception as ex:
            return str(ex)
        return ''

    def _check_response(self, resp):
        if resp.status_code >= 300:
            logger.warning(resp.content)
        resp.raise_for_status()
        return resp


class PicasaUploadConfig(QtWidgets.QWidget):
    delete_album = QtCore.pyqtSignal()
    new_album = QtCore.pyqtSignal()
    select_album = QtCore.pyqtSignal(six.text_type)
    update_album = QtCore.pyqtSignal(dict)

    def __init__(self, *arg, **kw):
        super(PicasaUploadConfig, self).__init__(*arg, **kw)
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.widgets = {}
        ## album details, left hand side
        album_group = QtWidgets.QGroupBox(self.tr('Collection / Album'))
        album_group.setLayout(QtWidgets.QHBoxLayout())
        album_form_left = QtWidgets.QFormLayout()
        album_form_left.setFieldGrowthPolicy(
            QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        album_group.layout().addLayout(album_form_left)
        # album title / selector
        self.albums = QtWidgets.QComboBox()
        self.albums.setEditable(True)
        self.albums.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
        self.albums.activated.connect(self.switch_album)
        self.albums.lineEdit().editingFinished.connect(self.new_title)
        album_form_left.addRow(self.tr('Title'), self.albums)
        # album description
        self.widgets['description'] = MultiLineEdit(spell_check=True)
        self.widgets['description'].editingFinished.connect(self.new_album_details)
        album_form_left.addRow(
            self.tr('Description'), self.widgets['description'])
        # album location
        self.widgets['location'] = QtWidgets.QLineEdit()
        self.widgets['location'].editingFinished.connect(self.new_album_details)
        album_form_left.addRow(self.tr('Place taken'), self.widgets['location'])
        # album visibility
        self.widgets['access'] = QtWidgets.QComboBox()
        self.widgets['access'].addItem(self.tr('Public on the web'), 'public')
        self.widgets['access'].addItem(
            self.tr('Limited, anyone with the link'), 'private')
        self.widgets['access'].addItem(self.tr('Only you'), 'protected')
        self.widgets['access'].currentIndexChanged.connect(self.new_access)
        album_form_left.addRow(self.tr('Visibility'), self.widgets['access'])
        ## album buttons
        buttons = QtWidgets.QHBoxLayout()
        buttons.addStretch(stretch=60)
        album_form_left.addRow(buttons)
        # new album
        new_album_button = QtWidgets.QPushButton(self.tr('New album'))
        new_album_button.clicked.connect(self.new_album)
        buttons.addWidget(new_album_button, stretch=20)
        # delete album
        delete_album_button = QtWidgets.QPushButton(self.tr('Delete album'))
        delete_album_button.clicked.connect(self.delete_album)
        buttons.addWidget(delete_album_button, stretch=20)
        ## album details, right hand side
        album_form_right = QtWidgets.QFormLayout()
        album_form_right.setFieldGrowthPolicy(
            QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        album_group.layout().addLayout(album_form_right)
        # album date
        self.widgets['timestamp'] = QtWidgets.QDateEdit()
        self.widgets['timestamp'].setMinimumDateTime(
            QtCore.QDateTime.fromTime_t(0))
        self.widgets['timestamp'].setCalendarPopup(True)
        self.widgets['timestamp'].editingFinished.connect(self.new_album_details)
        album_form_right.addRow(self.tr('Date'), self.widgets['timestamp'])
        # album thumbnail
        self.album_thumb = QtWidgets.QLabel()
        album_form_right.addRow(self.album_thumb)
        self.layout().addWidget(album_group)

    @QtCore.pyqtSlot()
    def new_title(self):
        value = self.albums.lineEdit().text()
        self.albums.setItemText(self.albums.currentIndex(), value)
        self.new_album_details()

    @QtCore.pyqtSlot(int)
    def new_access(self, index):
        self.new_album_details()

    @QtCore.pyqtSlot()
    def new_album_details(self):
        new_values = {
            'title'       : self.albums.itemText(self.albums.currentIndex()),
            'description' : self.widgets['description'].get_value(),
            'location'    : self.widgets['location'].text(),
            'access'      : self.widgets['access'].itemData(
                self.widgets['access'].currentIndex()),
            'timestamp'   : '{:d}'.format(
                self.widgets['timestamp'].dateTime().toTime_t() * 1000),
            }
        self.update_album.emit(new_values)

    @QtCore.pyqtSlot(int)
    def switch_album(self, index):
        self.select_album.emit(self.albums.itemData(index))

    def show_album(self, album):
        if album:
            self.albums.setCurrentIndex(self.albums.findData(album.id.text))
            self.widgets['description'].set_value(album.summary.text)
            self.widgets['location'].setText(album.location.text)
            self.widgets['access'].setCurrentIndex(
                self.widgets['access'].findData(album.access.text))
            self.widgets['timestamp'].setDateTime(
                QtCore.QDateTime.fromTime_t(int(album.timestamp.text) // 1000))
            if album.group.thumbnail is not None:
                QtWidgets.QApplication.processEvents()
                with Busy():
                    url = album.group.thumbnail[0].get('url')
                    image = QtGui.QPixmap()
                    image.loadFromData(urlopen(url).read())
                    self.album_thumb.setPixmap(image)
        else:
            self.widgets['description'].set_value(None)
            self.widgets['location'].clear()
            self.widgets['timestamp'].clear()
            self.album_thumb.clear()


class PicasaUploader(PhotiniUploader):
    session_factory = PicasaSession

    def __init__(self, *arg, **kw):
        config_store.remove_section('picasa')
        self.upload_config = PicasaUploadConfig()
        self.upload_config.delete_album.connect(self.delete_album)
        self.upload_config.new_album.connect(self.new_album)
        self.upload_config.select_album.connect(self.select_album)
        self.upload_config.update_album.connect(self.update_album)
        super(PicasaUploader, self).__init__(self.upload_config, *arg, **kw)
        self.service_name = self.tr('Google Photos')
        self.convert = {
            'types'   : ('bmp', 'gif', 'jpeg', 'png'),
            'msg'     : self.tr(
                'File "{0}" is of type "{1}", which Google Photos does not' +
                ' accept. Would you like to convert it to JPEG?'),
            'buttons' : QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Ignore,
            }
        # timer to store album data after it's edited
        self.album_changed = False
        self.timer = QtCore.QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(5000)
        self.timer.timeout.connect(self.save_changes)

    def get_albums(self, feed=None):
        if not feed:
            feed = self.session.get_feed()
        for album in feed.entry:
            if not album.get_link('edit'):
                # ignore 'system' albums
                continue
            yield album

    def load_user_data(self):
        self.current_album = None
        self.upload_config.albums.clear()
        if self.connected:
            feed = self.session.get_feed()
            self.show_user(feed.nickname.text, feed.thumbnail.text)
            for album in self.get_albums(feed):
                self.upload_config.albums.addItem(
                    album.title.text, album.id.text)
            self.set_current_album()
        else:
            self.show_user(None, None)
            self.upload_config.show_album(None)

    def get_upload_params(self):
        return self.current_album

    def upload_finished(self):
        # reload current album metadata (to update thumbnail)
        with Busy():
            self.set_current_album(self.current_album.id.text)

    @QtCore.pyqtSlot()
    def new_album(self):
        with Busy():
            self.save_changes()
            if not self.session.permitted('write'):
                self.refresh()
                return
            self.current_album = None
            self.upload_config.show_album(None)
            QtWidgets.QApplication.processEvents()
            album = PicasaNode()
            album.title.text = self.tr('New album')
            album.category.set('scheme', nsmap['gd'] + '#kind')
            album.category.set('term', nsmap['gphoto'] + '#album')
            album = self.session.new_album(album)
            self.upload_config.albums.addItem(album.title.text, album.id.text)
            self.set_current_album(album.id.text)

    @QtCore.pyqtSlot()
    def delete_album(self):
        album = self.current_album
        if int(album.numphotos.text) > 0:
            if QtWidgets.QMessageBox.question(
                self, self.tr('Delete album'),
                self.tr("""Are you sure you want to delete the album "{0}"?
Doing so will remove the album and its photos from all Google products."""
                        ).format(album.title.text),
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
                QtWidgets.QMessageBox.Cancel
                ) == QtWidgets.QMessageBox.Cancel:
                return
        with Busy():
            self.timer.stop()
            self.album_changed = False
            if not self.session.permitted('write'):
                self.refresh()
                return
            self.current_album = None
            self.upload_config.show_album(None)
            QtWidgets.QApplication.processEvents()
            self.session.delete_album(album)
            self.upload_config.albums.removeItem(
                self.upload_config.albums.findData(album.id.text))
            self.set_current_album()

    @QtCore.pyqtSlot(six.text_type)
    def select_album(self, album_id):
        with Busy():
            self.save_changes()
            if not self.session.permitted('read'):
                self.refresh()
                return
            self.current_album = None
            self.upload_config.show_album(None)
            QtWidgets.QApplication.processEvents()
            self.set_current_album(album_id)

    @QtCore.pyqtSlot(dict)
    def update_album(self, new_values):
        if not self.current_album:
            return
        if new_values['title'] != self.current_album.title.text:
            self.current_album.title.text = new_values['title']
            self.album_changed = True
        if new_values['description'] != (self.current_album.summary.text or ''):
            self.current_album.summary.text = new_values['description']
            self.album_changed = True
        if new_values['location'] != (self.current_album.location.text or ''):
            self.current_album.location.text = new_values['location']
            self.album_changed = True
        if new_values['access'] != self.current_album.access.text:
            self.current_album.access.text = new_values['access']
            self.album_changed = True
        if new_values['timestamp'] != self.current_album.timestamp.text:
            self.current_album.timestamp.text = new_values['timestamp']
            self.album_changed = True
        if self.album_changed:
            self.timer.start()

    def set_current_album(self, album_id=None):
        if self.upload_config.albums.count() == 0:
            self.new_album()
            return
        if album_id is None:
            album_id = self.upload_config.albums.itemData(0)
        self.current_album = None
        for album in self.get_albums():
            if album.id.text == album_id:
                self.upload_config.show_album(album)
                self.current_album = album
                return
        self.upload_config.show_album(None)

    def shutdown(self):
        self.save_changes()
        super(PicasaUploader, self).shutdown()

    def save_changes(self):
        self.timer.stop()
        if not self.album_changed:
            return
        self.album_changed = False
        with Busy():
            self.session.edit_node(self.current_album)
            self.set_current_album(self.current_album.id.text)



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-16  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import codecs
from datetime import datetime
from fractions import Fraction
import locale
import logging
import os

try:
    import pgi
    pgi.install_as_gi()
    using_pgi = True
except ImportError:
    using_pgi = False
import gi
for gexiv2_vsn in ('0.10', '0.4'):
    try:
        gi.require_version('GExiv2', gexiv2_vsn)
        break
    except ValueError:
        pass
from gi.repository import GObject, GExiv2
import six

from .pyqt import QtCore
from . import __version__

logger = logging.getLogger(__name__)

gexiv2_version = '{} {}, GExiv2 {}, GObject {}'.format(
    ('PyGI', 'pgi')[using_pgi], gi.__version__,
    GExiv2._version, GObject._version)

# pydoc gi.repository.GExiv2.Metadata is useful to see methods available

GExiv2.log_set_level(GExiv2.LogLevel.MUTE)

@six.python_2_unicode_compatible
class MetadataValue(object):
    # base for classes that store a metadata value, e.g. a string, int
    # or float
    def __init__(self, value):
        assert(value is not None)
        self.value = value

    @classmethod
    def from_exif(cls, file_value):
        return cls(file_value)

    def to_exif(self):
        return self.value

    @classmethod
    def from_iptc(cls, file_value):
        return cls(file_value)

    def to_iptc(self):
        return self.value

    @classmethod
    def from_xmp(cls, file_value):
        return cls(file_value)

    def to_xmp(self):
        return self.value

    def __str__(self):
        return six.text_type(self.value)

    def __nonzero__(self):
        return bool(self.value)

    def __eq__(self, other):
        return isinstance(other, MetadataValue) and self.value == other.value

    def __ne__(self, other):
        return not isinstance(other, MetadataValue) or self.value != other.value

    def contains(self, other):
        # "contains" = no need to merge or replace.
        return (not other) or (other.value == self.value)

    def merge(self, other, family=None):
        # Only called if contains returns False. Returns True if other
        # merged into self, False if self should be replaced by other.
        return False


class MetadataDictValue(MetadataValue):
    # base for classes that store a dictionary of metadata values, e.g.
    # latitude & longitude
    value = {}

    def __getattr__(self, name):
        if name in self.value:
            return self.value[name]
        return super(MetadataDictValue, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name in self.value:
            self.value[name] = value
            return
        super(MetadataDictValue, self).__setattr__(name, value)


class FocalLength(MetadataDictValue):
    # store actual focal length and 35mm film equivalent
    def __init__(self, value):
        if isinstance(value, six.string_types):
            value = value.split(',')
        fl, fl_35 = value
        if fl in (None, ''):
            fl = None
        else:
            fl = Fraction(fl).limit_denominator(1000000)
        if fl_35 in (None, ''):
            fl_35 = None
        else:
            fl_35 = int(fl_35)
        super(FocalLength, self).__init__({'fl': fl, 'fl_35': fl_35})

    @classmethod
    def from_exif(cls, file_value):
        focal_length, focal_length_35mm = file_value
        return cls((focal_length, focal_length_35mm))

    def to_exif(self):
        if self.fl is None:
            focal_length = None
        else:
            focal_length = '{:d}/{:d}'.format(
                self.fl.numerator, self.fl.denominator)
        if self.fl_35 is None:
            focal_length_35mm = None
        else:
            focal_length_35mm = '{:d}'.format(self.fl_35)
        return focal_length, focal_length_35mm

    @classmethod
    def from_xmp(cls, file_value):
        focal_length, focal_length_35mm = file_value
        return cls((focal_length, focal_length_35mm))

    def to_35(self, value):
        if self.fl and self.fl_35:
            return int((float(value) * self.fl_35 / self.fl) + 0.5)
        return None

    def from_35(self, value):
        if self.fl and self.fl_35:
            return round(float(value) * self.fl / self.fl_35, 2)
        return None

    def __nonzero__(self):
        return (self.fl is not None) or (self.fl_35 is not None)

    def contains(self, other):
        return (not other) or ((other.fl in (None, self.fl)) and
                               (other.fl_35 in (None, self.fl_35)))

    def merge(self, other, family=None):
        if self.fl is None:
            self.fl = other.fl
        if self.fl_35 is None:
            self.fl_35 = other.fl_35
        return True


class LatLon(MetadataDictValue):
    # simple class to store latitude and longitude
    def __init__(self, value):
        if isinstance(value, six.string_types):
            value = value.split(',')
        lat, lon = value
        super(LatLon, self).__init__({
            'lat' : round(float(lat), 6),
            'lon' : round(float(lon), 6),
            })

    @staticmethod
    def from_exif_part(value, ref):
        parts = [float(Fraction(x)) for x in value.split()] + [0.0, 0.0]
        result = parts[0] + (parts[1] / 60.0) + (parts[2] / 3600.0)
        if ref in ('S', 'W'):
            result = -result
        return result

    @classmethod
    def from_exif(cls, file_value):
        if not all(file_value):
            return None
        lat_string, lat_ref, lon_string, lon_ref = file_value
        return cls((cls.from_exif_part(lat_string, lat_ref),
                    cls.from_exif_part(lon_string, lon_ref)))

    @staticmethod
    def to_exif_part(value):
        if value >= 0.0:
            negative = False
        else:
            negative = True
            value = -value
        degrees = int(value)
        value = (value - degrees) * 60.0
        minutes = int(value)
        seconds = (value - minutes) * 60.0
        seconds = Fraction(seconds).limit_denominator(1000000)
        return '{:d}/1 {:d}/1 {:d}/{:d}'.format(
            degrees, minutes, seconds.numerator, seconds.denominator), negative

    def to_exif(self):
        lat_string, negative = self.to_exif_part(self.lat)
        lat_ref = 'NS'[negative]
        lon_string, negative = self.to_exif_part(self.lon)
        lon_ref = 'EW'[negative]
        return lat_string, lat_ref, lon_string, lon_ref

    @staticmethod
    def from_xmp_part(value):
        degrees, minutes = value[:-1].split(',')
        ref = value[-1]
        value = float(degrees) + (float(minutes) / 60.0)
        if ref in ('S', 'W'):
            value = -value
        return value

    @classmethod
    def from_xmp(cls, file_value):
        if not all(file_value):
            return None
        lat_string, lon_string = file_value
        return cls((cls.from_xmp_part(lat_string), cls.from_xmp_part(lon_string)))

    def __str__(self):
        return '{:.6f}, {:.6f}'.format(self.lat, self.lon)

    def contains(self, other):
        return (not other) or ((abs(other.lat - self.lat) < 0.000001) and
                               (abs(other.lon - self.lon) < 0.000001))


class LensSpec(MetadataDictValue):
    # simple class to store lens "specificaton"
    def __init__(self, value):
        if isinstance(value, six.string_types):
            sep = None
            if ',' in value:
                sep = ','
            value = value.split(sep)
        min_fl, max_fl, min_fl_fn, max_fl_fn = value
        super(LensSpec, self).__init__({
            'min_fl'    : Fraction(min_fl).limit_denominator(1000000),
            'max_fl'    : Fraction(max_fl).limit_denominator(1000000),
            'min_fl_fn' : Fraction(min_fl_fn).limit_denominator(1000000),
            'max_fl_fn' : Fraction(max_fl_fn).limit_denominator(1000000),
            })

    def __str__(self):
        return '{:g} {:g} {:g} {:g}'.format(
            float(self.min_fl),    float(self.max_fl),
            float(self.min_fl_fn), float(self.max_fl_fn))

    def to_exif(self):
        return ' '.join(
            ['{:d}/{:d}'.format(x.numerator, x.denominator) for x in (
                self.min_fl, self.max_fl, self.min_fl_fn, self.max_fl_fn)])


class DateTime(MetadataDictValue):
    # store date and time with "precision" to store how much is valid
    # tz_offset is stored in minutes
    def __init__(self, value):
        date_time, precision, tz_offset = value
        if date_time is None:
            # use a well known 'zero'
            date_time = datetime(1970, 1, 1)
        else:
            parts = [date_time.year, date_time.month, date_time.day,
                     date_time.hour, date_time.minute, date_time.second,
                     date_time.microsecond][:precision]
            while len(parts) < 3:
                parts.append(1)
            date_time = datetime(*parts)
        if precision <= 3:
            tz_offset = None
        super(DateTime, self).__init__({
            'datetime'  : date_time,
            'precision' : precision,
            'tz_offset' : tz_offset,
            })

    @classmethod
    def from_ISO_8601(cls, date_string, time_string, tz_string):
        """Sufficiently general ISO 8601 parser.

        Inputs must be in "basic" format, i.e. no '-' or ':' separators.
        See https://en.wikipedia.org/wiki/ISO_8601

        """
        # parse tz_string
        if tz_string:
            tz_offset = (int(tz_string[1:3]) * 60) + int(tz_string[3:])
            if tz_string[0] == '-':
                tz_offset = -tz_offset
        else:
            tz_offset = None
        if time_string == '000000':
            # assume no time information
            time_string = ''
            tz_offset = None
        datetime_string = date_string + time_string
        precision = min((len(datetime_string) - 2) // 2, 7)
        if precision <= 0:
            return None
        fmt = ''.join(('%Y', '%m', '%d', '%H', '%M', '%S', '.%f')[:precision])
        return cls(
            (datetime.strptime(datetime_string, fmt), precision, tz_offset))

    def to_ISO_8601(self, fmt=('%Y', '-%m', '-%d', 'T%H', ':%M', ':%S', '.%f'),
                    precision=None, time_zone=True):
        if precision is None:
            precision = self.precision
        fmt = ''.join(fmt[:precision])
        datetime_string = self.datetime.strftime(fmt)
        if precision > 3 and time_zone and self.tz_offset is not None:
            # add time zone
            minutes = self.tz_offset
            if minutes >= 0:
                datetime_string += '+'
            else:
                datetime_string += '-'
                minutes = -minutes
            datetime_string += '{:02d}:{:02d}'.format(
                minutes // 60, minutes % 60)
        return datetime_string

    # Exif datetime is always full resolution and valid. Assume a time
    # of 00:00:00 is a none value though.
    @classmethod
    def from_exif(cls, file_value):
        datetime_string, sub_sec_string = file_value
        if not datetime_string:
            return None
        # separate date & time and remove separators
        date_string = datetime_string[:10].replace(':', '')
        time_string = datetime_string[11:].replace(':', '')
        # append sub seconds
        if sub_sec_string:
            time_string += '.' + sub_sec_string
        return cls.from_ISO_8601(date_string, time_string, '')

    def to_exif(self):
        datetime_string, sep, sub_sec_string = self.to_ISO_8601(
            fmt=('%Y', ':%m', ':%d', ' %H', ':%M', ':%S', '.%f'),
            time_zone=False).partition('.')
        # pad out any missing values
        #                   YYYY mm dd HH MM SS
        datetime_string += '0000:01:01 00:00:00'[len(datetime_string):]
        return datetime_string, sub_sec_string

    # IPTC date & time should have no separators and be 8 and 11 chars
    # respectively (time includes time zone offset). I suspect the exiv2
    # library is adding separators, but am not sure.

    # The date (and time?) can have missing values represented by 00
    # according to
    # https://de.wikipedia.org/wiki/IPTC-IIM-Standard#IPTC-Felder
    @classmethod
    def from_iptc(cls, file_value):
        date_string, time_string = file_value
        if not date_string:
            return None
        # remove separators (that shouldn't be there)
        date_string = date_string.replace('-', '')
        # remove missing values
        while len(date_string) > 4 and date_string[-2:] == '00':
            date_string = date_string[:-2]
        if date_string == '0000':
            return None
        # ignore time if date is not full precision
        if len(date_string) < 8:
            time_string = ''
        if time_string:
            # remove separators (that shouldn't be there)
            time_string = time_string.replace(':', '')
            # split off time zone
            tz_string = time_string[6:]
            time_string = time_string[:6]
        else:
            tz_string = ''
            time_string = ''
        return cls.from_ISO_8601(date_string, time_string, tz_string)

    def to_iptc(self):
        if self.precision <= 3:
            date_string = self.to_ISO_8601()
            #               YYYY mm dd
            date_string += '0000-00-00'[len(date_string):]
            time_string = None
        else:
            datetime_string = self.to_ISO_8601(precision=6)
            date_string = datetime_string[:10]
            time_string = datetime_string[11:]
        return date_string, time_string

    # XMP uses extended ISO 8601, but the time cannot be hours only. See
    # p75 of
    # https://partners.adobe.com/public/developer/en/xmp/sdk/XMPspecification.pdf
    # According to p71, when converting Exif values with no time zone,
    # local time zone should be assumed. However, the MWG guidelines say
    # this must not be assumed to be the time zone where the photo is
    # processed. It also says the XMP standard has been revised to make
    # time zone information optional.
    @classmethod
    def from_xmp(cls, file_value):
        date_string, sep, time_string = file_value.partition('T')
        if len(time_string) > 6 and time_string[-6] in ('+', '-'):
            tz_string = time_string[-6:]
            time_string = time_string[:-6]
        elif len(time_string) > 1 and time_string[-1] == 'Z':
            tz_string = '+00:00'
            time_string = time_string[:-1]
        else:
            tz_string = ''
        return cls.from_ISO_8601(
            date_string.replace('-', ''), time_string.replace(':', ''),
            tz_string.replace(':', ''))

    def to_xmp(self):
        precision = self.precision
        if precision == 4:
            precision = 5
        return self.to_ISO_8601(precision=precision)

    def __str__(self):
        return self.to_ISO_8601()

    def contains(self, other):
        if (not other) or (other.value == self.value):
            return True
        if other.datetime != self.datetime:
            return False
        if other.precision < self.precision:
            return False
        return bool(other.tz_offset) == bool(self.tz_offset)

    def merge(self, other, family=None):
        result = False
        # some formats default to a higher precision
        if self.precision < 7 and self.precision > other.precision:
            self.precision = other.precision
            self.datetime = other.datetime
            result = True
        # don't trust IPTC time zone
        if self.tz_offset is None and family != 'Iptc':
            self.tz_offset = other.tz_offset
            result = True
        return result


@six.python_2_unicode_compatible
class MultiString(MetadataValue):
    def __init__(self, value):
        if isinstance(value, six.string_types):
            value = value.split(';')
        value = list(filter(bool, [x.strip() for x in value]))
        super(MultiString, self).__init__(value)

    def to_exif(self):
        return ';'.join(self.value)

    def __str__(self):
        return '; '.join(self.value)

    def contains(self, other):
        return (not other) or (not bool(
            [x for x in other.value if x not in self.value]))

    def merge(self, other, family=None):
        self.value += other.value
        return True


class String(MetadataValue):
    def __init__(self, value):
        if isinstance(value, list):
            value = value[0]
        super(String, self).__init__(six.text_type(value).strip())

    def contains(self, other):
        return (not other) or (other.value in self.value)

    def merge(self, other, family=None):
        self.value += ' // ' + other.value
        return True


class CharacterSet(String):
    known_encodings = {
        'ascii'   : '\x1b(B',
        'latin_1' : '\x1b/A',
        'latin1'  : '\x1b.A',
        'utf_8'   : '\x1b%G',
        }

    @classmethod
    def from_iptc(cls, file_value):
        for charset, encoding in cls.known_encodings.items():
            if encoding == file_value:
                return cls(charset)
        logger.warning('Unknown character encoding "%s"', repr(file_value))
        return None

    def to_iptc(self):
        return self.known_encodings[self.value]


class Software(String):
    @classmethod
    def from_iptc(cls, file_value):
        program, version = file_value
        if not program:
            return None
        if version:
            program += ' v' + version
        return cls(program)

    def to_iptc(self):
        return self.value.split(' v')


class Int(MetadataValue):
    def __init__(self, value):
        super(Int, self).__init__(int(value))

    def to_exif(self):
        return '{:d}'.format(self.value)

    def __nonzero__(self):
        return self.value is not None

    def __str__(self):
        return '{:d}'.format(self.value)


class Aperture(MetadataValue):
    def __init__(self, value):
        super(Aperture, self).__init__(Fraction(value))

    def to_exif(self):
        return '{:d}/{:d}'.format(self.value.numerator, self.value.denominator)

    def __nonzero__(self):
        return self.value is not None

    def __str__(self):
        return '{:g}'.format(float(self.value))

    def contains(self, other):
        return (not other) or ((min(other.value, self.value) /
                                max(other.value, self.value)) > 0.95)


# maximum length of Iptc data
_max_bytes = {
    'Iptc.Application2.Byline'           :   32,
    'Iptc.Application2.Caption'          : 2000,
    'Iptc.Application2.Copyright'        :  128,
    'Iptc.Application2.Headline'         :  256,
    'Iptc.Application2.Keywords'         :   64,
    'Iptc.Application2.ObjectName'       :   64,
    'Iptc.Application2.Program'          :   32,
    'Iptc.Application2.ProgramVersion'   :   10,
    'Iptc.Envelope.CharacterSet'         :   32,
    }

class MetadataHandler(GExiv2.Metadata):
    def __init__(self, path, image_data=None):
        super(MetadataHandler, self).__init__()
        self._logger = logging.getLogger(self.__class__.__name__)
        self._path = path
        # read metadata from file
        if image_data and not using_pgi:
            self.open_buf(image_data)
        else:
            self.open_path(self._path)
        # make list of possible character encodings
        self._encodings = []
        for name in ('utf_8', 'latin_1'):
            self._encodings.append(codecs.lookup(name).name)
        char_set = locale.getdefaultlocale()[1]
        if char_set:
            try:
                name = codecs.lookup(char_set).name
                if name not in self._encodings:
                    self._encodings.append(name)
            except LookupError:
                pass
        # convert IPTC data to UTF-8
        if not self.has_iptc():
            return
        current_encoding = self.get_value(
            CharacterSet, 'Iptc.Envelope.CharacterSet')
        if current_encoding:
            if current_encoding.value == 'utf_8':
                return
            try:
                name = codecs.lookup(current_encoding.value).name
                if name not in self._encodings:
                    self._encodings.insert(0, name)
            except LookupError:
                pass
        for tag in self.get_iptc_tags():
            try:
                value_list = self.get_tag_multiple(tag)
                self.set_tag_multiple(tag, value_list)
            except Exception as ex:
                self._logger.exception(ex)

    def _decode_string(self, value):
        if not value:
            return value
        for encoding in self._encodings:
            try:
                return value.decode(encoding)
            except UnicodeDecodeError:
                continue
        return value.decode('utf_8', 'replace')

    def get_value(self, data_type, tag):
        # get string or multiple strings
        if tag in ('Iptc.Application2.Byline', 'Iptc.Application2.Keywords',
                   'Xmp.dc.creator', 'Xmp.dc.subject'):
            file_value = self.get_tag_multiple(tag)
        elif tag in ('Xmp.dc.description', 'Xmp.dc.rights', 'Xmp.dc.title',
                     'Xmp.tiff.Copyright'):
            file_value = self.get_tag_multiple(tag)
            if file_value:
                file_value = file_value[0]
        else:
            file_value = self.get_tag_string(tag)
        if file_value is None or not any(file_value):
            return None
        # manipulate some tags' data
        if tag in ('Exif.Image.ApertureValue',
                   'Exif.Photo.ApertureValue'):
            # convert APEX aperture value
            file_value = 2.0 ** (Fraction(file_value) / 2.0)
        elif tag in ('Exif.Image.XPAuthor', 'Exif.Image.XPComment',
                     'Exif.Image.XPKeywords', 'Exif.Image.XPSubject',
                     'Exif.Image.XPTitle'):
            # decode UCS2 string
            file_value = bytearray(map(int, file_value.split()))
            file_value = file_value.decode('utf_16').strip('\x00')
        # convert to Photini data type
        if MetadataHandler.is_exif_tag(tag):
            return data_type.from_exif(file_value)
        if MetadataHandler.is_iptc_tag(tag):
            return data_type.from_iptc(file_value)
        return data_type.from_xmp(file_value)

    def clear_tag(self, tag):
        if isinstance(tag, tuple):
            for sub_tag in tag:
                super(MetadataHandler, self).clear_tag(sub_tag)
            return
        super(MetadataHandler, self).clear_tag(tag)

    def set_value(self, tag, value):
        if not value:
            self.clear_tag(tag)
            return
        # convert from Photini data type to string or multiple string
        if MetadataHandler.is_exif_tag(tag):
            file_value = value.to_exif()
        elif MetadataHandler.is_iptc_tag(tag):
            file_value = value.to_iptc()
        else:
            file_value = value.to_xmp()
        # write to file
        if tag in ('Iptc.Application2.Byline', 'Iptc.Application2.Keywords',
                   'Xmp.dc.creator', 'Xmp.dc.subject'):
            self.set_tag_multiple(tag, file_value)
        else:
            self.set_tag_string(tag, file_value)

    def get_tag_string(self, tag):
        if isinstance(tag, tuple):
            return list(map(self.get_tag_string, tag))
        try:
            result = super(MetadataHandler, self).get_tag_string(tag)
            if six.PY2:
                result = self._decode_string(result)
        except UnicodeDecodeError as ex:
            self._logger.error(str(ex))
            return ''
        return result

    def get_tag_multiple(self, tag):
        if isinstance(tag, tuple):
            return list(map(self.get_tag_multiple, tag))
        try:
            result = super(MetadataHandler, self).get_tag_multiple(tag)
            if six.PY2:
                result = list(map(self._decode_string, result))
        except UnicodeDecodeError as ex:
            self._logger.error(str(ex))
            return []
        return result

    def set_tag_string(self, tag, value):
        if isinstance(tag, tuple):
            for sub_tag, sub_value in zip(tag, value):
                self.set_tag_string(sub_tag, sub_value)
            return
        if not value:
            super(MetadataHandler, self).clear_tag(tag)
            return
        if MetadataHandler.is_iptc_tag(tag) and tag in _max_bytes:
            value = value.encode('utf_8')[:_max_bytes[tag]]
            if not six.PY2:
                value = value.decode('utf_8')
        elif six.PY2:
            value = value.encode('utf_8')
        super(MetadataHandler, self).set_tag_string(tag, value)

    def set_tag_multiple(self, tag, value):
        if isinstance(tag, tuple):
            for sub_tag, sub_value in zip(tag, value):
                self.set_tag_multiple(sub_tag, sub_value)
            return
        if not value:
            super(MetadataHandler, self).clear_tag(tag)
            return
        if MetadataHandler.is_iptc_tag(tag) and tag in _max_bytes:
            value = [x.encode('utf_8')[:_max_bytes[tag]] for x in value]
            if not six.PY2:
                value = [x.decode('utf_8') for x in value]
        elif six.PY2:
            value = [x.encode('utf_8') for x in value]
        super(MetadataHandler, self).set_tag_multiple(tag, value)

    @staticmethod
    def is_exif_tag(tag):
        if isinstance(tag, tuple):
            tag = tag[0]
        return GExiv2.Metadata.is_exif_tag(tag)

    @staticmethod
    def is_iptc_tag(tag):
        if isinstance(tag, tuple):
            tag = tag[0]
        return GExiv2.Metadata.is_iptc_tag(tag)

    def save(self):
        try:
            self.save_file(self._path)
        except GObject.GError as ex:
            self._logger.exception(ex)
            return False
        return True

    def copy(self, other, exif=True, iptc=True, xmp=True):
        # copy from other to self
        if exif:
            for tag in other.get_exif_tags():
                self.set_tag_string(tag, other.get_tag_string(tag))
        if iptc:
            for tag in other.get_iptc_tags():
                self.set_tag_multiple(tag, other.get_tag_multiple(tag))
        if xmp:
            for tag in other.get_xmp_tags():
                self.set_tag_multiple(tag, other.get_tag_multiple(tag))


class Metadata(QtCore.QObject):
    # type of each Photini data field's data
    _data_type = {
        'aperture'       : Aperture,
        'camera_model'   : String,
        'character_set'  : CharacterSet,
        'copyright'      : String,
        'creator'        : MultiString,
        'date_digitised' : DateTime,
        'date_modified'  : DateTime,
        'date_taken'     : DateTime,
        'description'    : String,
        'focal_length'   : FocalLength,
        'keywords'       : MultiString,
        'latlong'        : LatLon,
        'lens_make'      : String,
        'lens_model'     : String,
        'lens_serial'    : String,
        'lens_spec'      : LensSpec,
        'orientation'    : Int,
        'software'       : Software,
        'title'          : String,
        }
    # mapping of preferred tags to Photini data fields
    _primary_tags = {
        'aperture'       : {'Exif' : 'Exif.Photo.FNumber'},
        'camera_model'   : {'Exif' : 'Exif.Image.Model'},
        'character_set'  : {'Iptc' : 'Iptc.Envelope.CharacterSet'},
        'copyright'      : {'Exif' : 'Exif.Image.Copyright',
                            'Xmp'  : 'Xmp.dc.rights',
                            'Iptc' : 'Iptc.Application2.Copyright'},
        'creator'        : {'Exif' : 'Exif.Image.Artist',
                            'Xmp'  : 'Xmp.dc.creator',
                            'Iptc' : 'Iptc.Application2.Byline'},
        'date_digitised' : {'Exif' : ('Exif.Photo.DateTimeDigitized',
                                      'Exif.Photo.SubSecTimeDigitized'),
                            'Xmp'  : 'Xmp.xmp.CreateDate',
                            'Iptc' : ('Iptc.Application2.DigitizationDate',
                                      'Iptc.Application2.DigitizationTime')},
        'date_modified'  : {'Exif' : ('Exif.Image.DateTime',
                                      'Exif.Photo.SubSecTime'),
                            'Xmp'  : 'Xmp.xmp.ModifyDate'},
        'date_taken'     : {'Exif' : ('Exif.Photo.DateTimeOriginal',
                                      'Exif.Photo.SubSecTimeOriginal'),
                            'Xmp'  : 'Xmp.photoshop.DateCreated',
                            'Iptc' : ('Iptc.Application2.DateCreated',
                                      'Iptc.Application2.TimeCreated')},
        'description'    : {'Exif' : 'Exif.Image.ImageDescription',
                            'Xmp'  : 'Xmp.dc.description',
                            'Iptc' : 'Iptc.Application2.Caption'},
        'focal_length'   : {'Exif' : ('Exif.Photo.FocalLength',
                                      'Exif.Photo.FocalLengthIn35mmFilm')},
        'keywords'       : {'Xmp'  : 'Xmp.dc.subject',
                            'Iptc' : 'Iptc.Application2.Keywords'},
        'latlong'        : {'Exif' : ('Exif.GPSInfo.GPSLatitude',
                                      'Exif.GPSInfo.GPSLatitudeRef',
                                      'Exif.GPSInfo.GPSLongitude',
                                      'Exif.GPSInfo.GPSLongitudeRef')},
        'lens_make'      : {'Exif' : 'Exif.Photo.LensMake'},
        'lens_model'     : {'Exif' : 'Exif.Photo.LensModel'},
        'lens_serial'    : {'Exif' : 'Exif.Photo.LensSerialNumber'},
        'lens_spec'      : {'Exif' : 'Exif.Photo.LensSpecification'},
        'orientation'    : {'Exif' : 'Exif.Image.Orientation'},
        'software'       : {'Exif' : 'Exif.Image.ProcessingSoftware',
                            'Iptc' : ('Iptc.Application2.Program',
                                      'Iptc.Application2.ProgramVersion')},
        'title'          : {'Xmp'  : 'Xmp.dc.title',
                            'Iptc' : 'Iptc.Application2.ObjectName'},
        }
    # mapping of duplicate tags to Photini data fields
    # data in these is merged in when data is read
    # they get deleted when data is written
    _secondary_tags = {
        'aperture'       : {'Exif' : ('Exif.Image.FNumber',
                                      'Exif.Image.ApertureValue',
                                      'Exif.Photo.ApertureValue',),
                            'Xmp'  : ('Xmp.exif.FNumber',
                                      'Xmp.exif.ApertureValue')},
        'character_set'  : {},
        'camera_model'   : {'Exif' : ('Exif.Image.UniqueCameraModel',)},
        'copyright'      : {'Xmp'  : ('Xmp.tiff.Copyright',)},
        'creator'        : {'Exif' : ('Exif.Image.XPAuthor',),
                            'Xmp'  : ('Xmp.tiff.Artist',)},
        'date_digitised' : {'Xmp'  : ('Xmp.exif.DateTimeDigitized',)},
        'date_modified'  : {'Xmp'  : ('Xmp.tiff.DateTime',)},
        'date_taken'     : {'Exif' : (('Exif.Image.DateTimeOriginal',
                                       'NoData'),),
                            'Xmp'  : ('Xmp.exif.DateTimeOriginal',)},
        'description'    : {'Exif' : ('Exif.Image.XPComment',
                                      'Exif.Image.XPSubject'),
                            'Xmp'  : ('Xmp.tiff.ImageDescription',)},
        'focal_length'   : {'Exif' : (('Exif.Image.FocalLength',
                                       'Exif.Photo.FocalLengthIn35mmFilm'),),
                            'Xmp'  : (('Xmp.exif.FocalLength',
                                       'Xmp.exif.FocalLengthIn35mmFilm'),)},
        'keywords'       : {'Exif' : ('Exif.Image.XPKeywords',)},
        'latlong'        : {'Xmp'  : (('Xmp.exif.GPSLatitude',
                                       'Xmp.exif.GPSLongitude'),)},
        'lens_make'      : {},
        'lens_model'     : {},
        'lens_serial'    : {},
        'lens_spec'      : {},
        'orientation'    : {'Xmp'  : ('Xmp.tiff.Orientation',)},
        'software'       : {},
        'title'          : {'Exif' : ('Exif.Image.XPTitle',),
                            'Iptc' : ('Iptc.Application2.Headline',)},
        }
    # tags that aren't read but are cleared when Photini data is written
    _clear_tags = {
        'aperture'       : (),
        'character_set'  : (),
        'camera_model'   : (),
        'copyright'      : (),
        'creator'        : (),
        'date_digitised' : (),
        'date_modified'  : (),
        'date_taken'     : (),
        'description'    : (),
        'focal_length'   : (),
        'keywords'       : (),
        'latlong'        : (),
        'lens_make'      : (),
        'lens_model'     : ('Exif.Canon.LensModel', 'Exif.CanonCs.LensType'),
        'lens_serial'    : (),
        'lens_spec'      : ('Exif.CanonCs.Lens', 'Exif.CanonCs.MaxAperture',
                            'Exif.CanonCs.MinAperture', 'Exif.CanonCs.ShortFocal'),
        'orientation'    : (),
        'software'       : (),
        'title'          : (),
        }
    def __init__(self, path, image_data, parent=None):
        super(Metadata, self).__init__(parent)
        self.logger = logging.getLogger(self.__class__.__name__)
        # create metadata handlers for image file and/or sidecar
        self._path = path
        self._sc_path = self._find_side_car(path)
        if self._sc_path:
            self._sc = MetadataHandler(self._sc_path)
        else:
            self._sc = None
        try:
            self._if = MetadataHandler(path, image_data)
        except Exception:
            self._if = None
            if not self._sc:
                self.create_side_car()
        self._unsaved = False

    def _find_side_car(self, path):
        for base in (os.path.splitext(path)[0], path):
            for ext in ('.xmp', '.XMP'):
                result = base + ext
                if os.path.exists(result):
                    return result
        return None

    def create_side_car(self):
        self._sc_path = self._path + '.xmp'
        with open(self._sc_path, 'w') as of:
            of.write('<x:xmpmeta x:xmptk="XMP Core 4.4.0-Exiv2" ')
            of.write('xmlns:x="adobe:ns:meta/">\n')
            of.write('</x:xmpmeta>')
        self._sc = MetadataHandler(self._sc_path)
        if self._if:
            self._sc.copy(self._if)

    def save(self, if_mode, sc_mode, force_iptc):
        if not self._unsaved:
            return
        self.software = 'Photini editor v' + __version__
        self.character_set = 'utf_8'
        save_iptc = force_iptc or self.has_iptc()
        for name in self._primary_tags:
            value = getattr(self, name)
            # write data to primary tags
            for family in self._primary_tags[name]:
                if save_iptc or family != 'Iptc':
                    tag = self._primary_tags[name][family]
                    self.set_value(tag, value)
            # delete secondary tags
            for family in self._secondary_tags[name]:
                for tag in self._secondary_tags[name][family]:
                    self.set_value(tag, None)
            # clear duplicated but unreadable data
            for tag in self._clear_tags[name]:
                self.set_value(tag, None)
        if self._if and sc_mode == 'delete' and self._sc:
            self._if.copy(self._sc)
        OK = False
        if self._if and if_mode:
            OK = self._if.save()
        if sc_mode == 'delete' and self._sc and OK:
            os.unlink(self._sc_path)
            self._sc = None
        if sc_mode == 'auto' and not self._sc and not OK:
            self.create_side_car()
        if sc_mode == 'always' and not self._sc:
            self.create_side_car()
        if self._sc:
            OK = self._sc.save()
        self._set_unsaved(not OK)

    # getters: use sidecar if tag is present, otherwise use image file
    def get_value(self, data_type, tag):
        result = None
        if self._sc:
            result = self._sc.get_value(data_type, tag)
        if self._if and not result:
            result = self._if.get_value(data_type, tag)
        return result

    def has_iptc(self):
        if self._sc and self._sc.has_iptc():
            return True
        if self._if and self._if.has_iptc():
            return True
        return False

    # setters: set in both sidecar and image file
    def set_value(self, tag, value):
        if self._sc:
            self._sc.set_value(tag, value)
        if self._if:
            self._if.set_value(tag, value)

    def __getattr__(self, name):
        if name not in self._primary_tags:
            return super(Metadata, self).__getattr__(name)
        # get values from all 3 families
        value = {'Exif': None, 'Iptc': None, 'Xmp': None}
        used_tag = {'Exif': None, 'Iptc': None, 'Xmp': None}
        for family in self._primary_tags[name]:
            tag = self._primary_tags[name][family]
            try:
                value[family] = self.get_value(self._data_type[name], tag)
                used_tag[family] = tag
            except Exception as ex:
                self.logger.exception(ex)
        # merge conflicting data from secondary tags
        for family in self._secondary_tags[name]:
            for tag in self._secondary_tags[name][family]:
                try:
                    new_value = self.get_value(self._data_type[name], tag)
                except Exception as ex:
                    self.logger.exception(ex)
                    continue
                if not new_value:
                    continue
                elif not value[family]:
                    value[family] = new_value
                    used_tag[family] = tag
                elif value[family].contains(new_value):
                    continue
                elif value[family].merge(new_value):
                    self.logger.warning(
                        '%s: merged %s into %s',
                        os.path.basename(self._path), tag, used_tag[family])
                else:
                    self.logger.warning(
                        '%s: using %s value "%s", ignoring %s value "%s"',
                        os.path.basename(self._path), used_tag[family],
                        str(value[family]), tag, str(new_value))
        # choose preferred family
        if value['Exif']:
            preference = 'Exif'
        elif value['Xmp']:
            preference = 'Xmp'
        else:
            preference = 'Iptc'
        # merge in non-matching data so user can review it
        result = value[preference]
        if result:
            for family in ('Exif', 'Xmp', 'Iptc'):
                other = value[family]
                if result.contains(other):
                    continue
                elif result.merge(other, family):
                    self.logger.warning(
                        '%s: merged %s data into %s',
                        os.path.basename(self._path),
                        used_tag[family], used_tag[preference])
                else:
                    self.logger.warning(
                        '%s: using %s value "%s", ignoring %s value "%s"',
                        os.path.basename(self._path), used_tag[preference],
                        str(result), used_tag[family], str(other))
        # add value to object attributes so __getattr__ doesn't get
        # called again
        super(Metadata, self).__setattr__(name, result)
        return result

    def __setattr__(self, name, value):
        if name not in self._primary_tags:
            return super(Metadata, self).__setattr__(name, value)
        if value in (None, '', [], {}):
            value = None
        elif not isinstance(value, self._data_type[name]):
            value = self._data_type[name](value)
            if not value:
                value = None
        if getattr(self, name) == value:
            return
        super(Metadata, self).__setattr__(name, value)
        self._set_unsaved(True)

    new_status = QtCore.pyqtSignal(bool)
    def _set_unsaved(self, status):
        self._unsaved = status
        self.new_status.emit(self._unsaved)

    def changed(self):
        return self._unsaved



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-16  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

"""
usage: editor.py [options]
options are:
  -h       | --help        display this help
  -v       | --verbose     increase number of logging messages
  -V       | --version     display version information and exit
"""

from __future__ import unicode_literals

import six
import logging
from optparse import OptionParser
import os
import sys
from six.moves.urllib.request import getproxies
from six.moves.urllib.parse import urlparse
import webbrowser

import pkg_resources

from .configstore import config_store
from .bingmap import BingMap
from .descriptive import Descriptive
try:
    from .facebook import FacebookUploader
except ImportError:
    FacebookUploader = None
try:
    from .flickr import FlickrUploader
except ImportError:
    FlickrUploader = None
from .editsettings import EditSettings
from .googlemap import GoogleMap
from .importer import Importer
from .metadata import gexiv2_version
from .openstreetmap import OpenStreetMap
from .imagelist import ImageList
from .loggerwindow import LoggerWindow
try:
    from .picasa import PicasaUploader
except ImportError:
    PicasaUploader = None
from .pyqt import Qt, QtCore, QtGui, QNetworkProxy, QtWidgets, qt_version_info
from .spelling import enchant_version, SpellingManager
from .technical import Technical
from . import __version__, build

class QTabBar(QtWidgets.QTabBar):
    def tabSizeHint(self, index):
        size = super(QTabBar, self).tabSizeHint(index)
        size.setWidth(max(size.width(), 90))
        return size


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, verbose, initial_files):
        super(MainWindow, self).__init__()
        self.setWindowTitle(self.tr("Photini photo metadata editor"))
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(
            pkg_resources.resource_string('photini', 'data/icon_48.png'))
        icon = QtGui.QIcon(pixmap)
        self.setWindowIcon(icon)
        self.selection = list()
        # logger window
        self.loggerwindow = LoggerWindow(verbose)
        self.loggerwindow.setWindowIcon(icon)
        self.logger = logging.getLogger(self.__class__.__name__)
        # set network proxy
        proxies = getproxies()
        if 'http' in proxies:
            parsed = urlparse(proxies['http'])
            QNetworkProxy.setApplicationProxy(QNetworkProxy(
                QNetworkProxy.HttpProxy, parsed.hostname, parsed.port))
        # restore size
        size = self.width(), self.height()
        self.resize(*eval(config_store.get('main_window', 'size', str(size))))
        # image selector
        self.image_list = ImageList()
        self.image_list.selection_changed.connect(self.new_selection)
        self.image_list.new_metadata.connect(self.new_metadata)
        # spelling manager
        self.spelling_manager = SpellingManager()
        # prepare list of tabs and associated stuff
        self.tab_list = (
            {'name'  : self.tr('&Descriptive metadata'),
             'key'   : 'descriptive_metadata',
             'class' : Descriptive},
            {'name'  : self.tr('&Technical metadata'),
             'key'   : 'technical_metadata',
             'class' : Technical},
            {'name'  : self.tr('Map (&Google)'),
             'key'   : 'map_google',
             'class' : GoogleMap},
            {'name'  : self.tr('Map (&Bing)'),
             'key'   : 'map_bing',
             'class' : BingMap},
            {'name'  : self.tr('Map (&OSM)'),
             'key'   : 'map_osm',
             'class' : OpenStreetMap},
            {'name'  : self.tr('&Flickr upload'),
             'key'   : 'flickr_upload',
             'class' : FlickrUploader},
            {'name'  : self.tr('Google &Photos upload'),
             'key'   : 'picasa_upload',
             'class' : PicasaUploader},
            {'name'  : self.tr('Faceboo&k upload'),
             'key'   : 'facebook_upload',
             'class' : FacebookUploader},
            {'name'  : self.tr('&Import photos'),
             'key'   : 'import_photos',
             'class' : Importer},
            )
        for tab in self.tab_list:
            if tab['class']:
                tab['object'] = tab['class'](self.image_list)
            else:
                tab['object'] = None
        # file menu
        file_menu = self.menuBar().addMenu(self.tr('File'))
        open_action = QtWidgets.QAction(self.tr('Open images'), self)
        open_action.setShortcuts(QtGui.QKeySequence.Open)
        open_action.triggered.connect(self.image_list.open_files)
        file_menu.addAction(open_action)
        self.save_action = QtWidgets.QAction(
            self.tr('Save images with new data'), self)
        self.save_action.setShortcuts(QtGui.QKeySequence.Save)
        self.save_action.setEnabled(False)
        self.save_action.triggered.connect(self.image_list.save_files)
        file_menu.addAction(self.save_action)
        self.close_action = QtWidgets.QAction(
            self.tr('Close selected images'), self)
        self.close_action.setEnabled(False)
        self.close_action.triggered.connect(self.close_files)
        file_menu.addAction(self.close_action)
        close_all_action = QtWidgets.QAction(self.tr('Close all images'), self)
        close_all_action.triggered.connect(self.close_all_files)
        file_menu.addAction(close_all_action)
        file_menu.addSeparator()
        quit_action = QtWidgets.QAction(self.tr('Quit'), self)
        quit_action.setShortcuts(
            [QtGui.QKeySequence.Quit, QtGui.QKeySequence.Close])
        quit_action.triggered.connect(
            QtWidgets.QApplication.instance().closeAllWindows)
        file_menu.addAction(quit_action)
        # options menu
        options_menu = self.menuBar().addMenu(self.tr('Options'))
        settings_action = QtWidgets.QAction(self.tr('Settings'), self)
        settings_action.triggered.connect(self.edit_settings)
        options_menu.addAction(settings_action)
        options_menu.addSeparator()
        for tab in self.tab_list:
            name = tab['name'].replace('&', '')
            tab['action'] = QtWidgets.QAction(name, self)
            tab['action'].setCheckable(True)
            if tab['class']:
                tab['action'].setChecked(
                    eval(config_store.get('tabs', tab['key'], 'True')))
            else:
                tab['action'].setEnabled(False)
            tab['action'].triggered.connect(self.add_tabs)
            options_menu.addAction(tab['action'])
        # spelling menu
        languages = self.spelling_manager.available_languages()
        spelling_menu = self.menuBar().addMenu(self.tr('Spelling'))
        enable_action = QtWidgets.QAction(self.tr('Enable spell check'), self)
        enable_action.setEnabled(bool(languages))
        enable_action.setCheckable(True)
        enable_action.setChecked(self.spelling_manager.enabled())
        enable_action.toggled.connect(self.spelling_manager.enable)
        spelling_menu.addAction(enable_action)
        language_menu = QtWidgets.QMenu(self.tr('Choose language'), self)
        language_menu.setEnabled(bool(languages))
        language_group = QtWidgets.QActionGroup(self)
        current_language = self.spelling_manager.current_language()
        for tag in languages:
            language_action = QtWidgets.QAction(tag, self)
            language_action.setCheckable(True)
            language_action.setChecked(tag == current_language)
            language_action.setActionGroup(language_group)
            language_menu.addAction(language_action)
        language_group.triggered.connect(self.spelling_manager.set_language)
        spelling_menu.addMenu(language_menu)
        # help menu
        help_menu = self.menuBar().addMenu(self.tr('Help'))
        about_action = QtWidgets.QAction(self.tr('About Photini'), self)
        about_action.triggered.connect(self.about)
        help_menu.addAction(about_action)
        help_menu.addSeparator()
        help_action = QtWidgets.QAction(self.tr('Photini documentation'), self)
        help_action.triggered.connect(self.open_docs)
        help_menu.addAction(help_action)
        # main application area
        self.central_widget = QtWidgets.QSplitter()
        self.central_widget.setOrientation(Qt.Vertical)
        self.central_widget.setChildrenCollapsible(False)
        self.tabs = QtWidgets.QTabWidget()
        self.tabs.setTabBar(QTabBar())
        self.tabs.setElideMode(Qt.ElideRight)
        self.tabs.currentChanged.connect(self.new_tab)
        self.add_tabs()
        self.central_widget.addWidget(self.tabs)
        self.central_widget.addWidget(self.image_list)
        size = self.central_widget.sizes()
        self.central_widget.setSizes(eval(
            config_store.get('main_window', 'split', str(size))))
        self.central_widget.splitterMoved.connect(self.new_split)
        self.setCentralWidget(self.central_widget)
        # open files given on command line, after GUI is displayed
        self.initial_files = initial_files
        if self.initial_files:
            QtCore.QTimer.singleShot(0, self.open_initial_files)

    @QtCore.pyqtSlot()
    def open_initial_files(self):
        self.image_list.open_file_list(self.initial_files)

    def add_tabs(self):
        was_blocked = self.tabs.blockSignals(True)
        current = self.tabs.currentWidget()
        self.tabs.clear()
        for tab in self.tab_list:
            use_tab = tab['action'].isChecked()
            config_store.set('tabs', tab['key'], str(use_tab))
            if tab['object'] and use_tab:
                self.tabs.addTab(tab['object'], tab['name'])
        self.tabs.blockSignals(was_blocked)
        if current:
            self.tabs.setCurrentWidget(current)
        self.new_tab(-1)

    def open_docs(self):
        webbrowser.open_new('http://photini.readthedocs.io/')

    def close_files(self):
        self._close_files(False)

    def close_all_files(self):
        self._close_files(True)

    def _close_files(self, all_files):
        if self.image_list.unsaved_files_dialog(all_files=all_files):
            self.image_list.close_files(all_files)

    def closeEvent(self, event):
        for n in range(self.tabs.count()):
            if self.tabs.widget(n).do_not_close():
                event.ignore()
                return
        self.image_list.unsaved_files_dialog(all_files=True, with_cancel=False)
        for n in range(self.tabs.count()):
            self.tabs.widget(n).shutdown()
        config_store.shutdown()
        self.loggerwindow.shutdown()
        super(MainWindow, self).closeEvent(event)

    def edit_settings(self):
        dialog = EditSettings(self)
        dialog.exec_()
        self.tabs.currentWidget().refresh()

    @QtCore.pyqtSlot()
    def about(self):
        text = self.tr("""
<table width="100%"><tr>
<td align="center" width="70%">
<h1>Photini</h1>
<h3>version {0}</h3>
<h4>build {1}</h4>
</td>
<td align="center"><img src="{2}" /></td>
</tr></table>
<p>&copy; Jim Easterbrook <a href="mailto:jim@jim-easterbrook.me.uk">
jim@jim-easterbrook.me.uk</a><br /><br />
An easy to use digital photograph metadata editor.<br />
Open source package available from
<a href="https://github.com/jim-easterbrook/Photini">
github.com/jim-easterbrook/Photini</a>.</p>
<p>This program is released with a GNU General Public License. For
details click the 'show details' button.</p>
""").format(__version__, build,
            pkg_resources.resource_filename('photini', 'data/icon_120.png'))
        dialog = QtWidgets.QMessageBox(self)
        dialog.setWindowTitle(self.tr('Photini: about'))
        dialog.setText(text)
        licence = pkg_resources.resource_string('photini', 'data/LICENSE.txt')
        dialog.setDetailedText(licence.decode('utf-8'))
        dialog.exec_()

    @QtCore.pyqtSlot(int, int)
    def new_split(self, pos, index):
        config_store.set(
            'main_window', 'split', str(self.central_widget.sizes()))

    @QtCore.pyqtSlot(int)
    def new_tab(self, index):
        current = self.tabs.currentWidget()
        if current:
            self.image_list.set_drag_to_map(None)
            current.refresh()
            self.image_list.emit_selection()

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        self.close_action.setEnabled(len(selection) > 0)
        self.tabs.currentWidget().new_selection(selection)

    @QtCore.pyqtSlot(bool)
    def new_metadata(self, unsaved_data):
        self.save_action.setEnabled(unsaved_data)

    def resizeEvent(self, event):
        size = self.width(), self.height()
        config_store.set('main_window', 'size', str(size))

def main(argv=None):
    if argv:
        sys.argv = argv
    # let PyQt handle its options (need at least one argument after options)
    sys.argv.append('xxx')
    app = QtWidgets.QApplication(sys.argv)
    del sys.argv[-1]
    # install translation
    if qt_version_info < (5, 0):
        QtCore.QTextCodec.setCodecForTr(QtCore.QTextCodec.codecForName('utf-8'))
    locale = QtCore.QLocale.system()
    translator = QtCore.QTranslator()
    translator.load(locale, 'photini', '.', pkg_resources.resource_filename(
        'photini', 'data/lang'), '.qm')
    app.installTranslator(translator)
    qt_translator = QtCore.QTranslator()
    qt_translator.load(
        locale, 'qt', '_',
        QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.TranslationsPath))
    app.installTranslator(qt_translator)
    # parse remaining arguments
    version = 'Photini ' + __version__ + ', build ' + build
    version += '\n  Python ' + sys.version
    version += '\n  ' + gexiv2_version
    version += '\n  PyQt {}, Qt {}'.format(
        QtCore.PYQT_VERSION_STR, QtCore.QT_VERSION_STR)
    if enchant_version:
        version += '\n  ' + enchant_version
    if FlickrUploader:
        from .flickr import flickr_version
        version += '\n  ' + flickr_version
    parser = OptionParser(
        usage=six.text_type(QtCore.QCoreApplication.translate(
            'main', 'Usage: %prog [options] [file_name, ...]')),
        version=version,
        description=six.text_type(QtCore.QCoreApplication.translate(
            'main', 'Photini photo metadata editor')))
    parser.add_option(
        '-v', '--verbose', action='count', default=0,
        help=six.text_type(QtCore.QCoreApplication.translate(
            'main', 'increase number of logging messages')))
    options, args = parser.parse_args()
    # create GUI and run application event loop
    main = MainWindow(options.verbose, args)
    main.show()
    return app.exec_()

if __name__ == "__main__":
    sys.exit(main())
